/**

	ATMega and TSIC temperature measurement

	Written and maintained by Matthias Pueski

	Copyright (c) 2013 Matthias Pueski

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/**
 *	Title   : Measuring temperature with the ATMega8 and a TSIC 0306
 *	          This version is interrrupt driven
 *	Author  : Matthias Pueski
 *	File    : main.c
 *	Hardware: ATMega8/16MHz
 *
 *	DESCRIPTION :
 *
 *	This program measures the temperature of a TSIC0306 temperature sensor, which is
 *	connected to the INT1 pin. The sensor itself is powered via the ZACPIN defined
 *	in zacwire.h
 *
 *	At the beginning the TSIC is powered on and when the first falling edge occurs the measurement
 *	starts. Timer 1 is started in normal mode. When the first rising edge occurs the timer is being
 *	stopped and the value from the TCNT1 register is taken as strobe value.
 *	From now on on each falling edge of the INT1 pin the timer1 is started in ctc mode.	Each time
 *	the OCR1A register is equal to the timer value, the TIMER1_COMPA_vect interrupt	service routine
 *	is being triggered. This is exactly the 50% duty cycle time which is needed to detect the state
 *	of each bit.
 *	After measuring the MSB the parity bit and the stop bit are being skipped, then the measurement
 *	for the LSB continues, at the end the parity bit is skipped again and the measurement is complete.
 *	The whole thing is driven by a "state machine" like mechanism. This in fact not a real state machine,
 *	since we have no events, rules or transitions.
 *
 *  This code doesn't represent the most efficient solution. The goal was to understand the zacwire protocol
 *	and to write human readable code. Feel free to improve the code and give suggestions.
 *
 *	Author : Matthias Pueski (pueski@gmx.de)
 *
 */
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "zacwire.h"
#include "lcd.h"
#include "int_ctrl.h"
#include "timer_ctrl.h"
// initial state of the reading "state machine"
zacstate state = ZAC_READING_START_BIT_LOW;
// half the value of tstrobe
volatile uint16_t strobe = 0;
// high byte of a zacwire packet
volatile uint8_t zac_high_byte = 0;
// low byte of a zacwire packet
volatile uint8_t zac_low_byte = 0;
// the bit number of the byte currently being read including parity
volatile uint8_t zac_current_bit = 8;
// output buffer
volatile uint8_t buffer[2];

// interrupt service routine for the timer1 ctc
ISR (TIMER1_COMPA_vect) {

	// stop the counter and reset timer count
	timer1_stop_ctc();
	TCNT1 = 0;

	// reading the HIGH byte of a zacwire packet
	if (state == ZAC_READING_HIGH_BYTE) {
		// read bits ignoring the last bit (parity)
		// actually reading 9 bits in total
		if (PIND & (1 << PIND3)) {
			if (zac_current_bit > 0) {
				zac_high_byte |= (1 << (zac_current_bit - 1));
			}
		}
		// decrement current bit every trigger
		if (zac_current_bit > 0) {
			zac_current_bit--;
		}
		// have read all bits from HIGH byte, switch
		// to LOW byte
		else {
			state = ZAC_READING_LOW_BYTE;
			// now we have one bit more since there comes another stop bit
			zac_current_bit = 9;
		}
	}

	// after reading the parity bit of the first byte, we can omit
	// the stop bit, since the next falling edge does not
	// happen until the next start bit (LOW byte)

	// reading the LOW byte of a zacwire packet
	else if (state == ZAC_READING_LOW_BYTE) {
		// read bits ignoring the first bit (start bit) and the last bit (parity)
		// actually reading 10 bits in total
		if (PIND & (1 << PIND3)) {
			if (zac_current_bit > 0 && zac_current_bit < 9) {
				zac_low_byte |= (1 << (zac_current_bit - 1));
			}
		}
		// decrement current bit every trigger
		if (zac_current_bit > 0) {
			zac_current_bit--;
		}
		// all bits read => back to the beginning
		else {
			zac_current_bit = 8;
			state = ZAC_READING_START_BIT_LOW;
		}
	}
}

// Interrupt service routine for the int1 pin
ISR (INT1_vect) {
	/**
	 * Beginning of zacwire packet, first falling edge
	 * start measuring tstrobe time
	 */
	if (state == ZAC_READING_START_BIT_LOW) {
		// if there was on measurement cycle, store data
		// from zacwire in usb transfer buffer
		if (zac_high_byte > 0)
			buffer[0] = zac_high_byte;
		if (zac_low_byte > 0)
			buffer[1] = zac_low_byte;
		// reset measured values
		zac_high_byte = 0;
		zac_low_byte = 0;
		// start timer1 and use rising edge for int1
		// if the next rising edge happens this is the second half of the start bit
		int1_switch_rising_edge();
		timer1_start_normal();
		state = ZAC_READING_START_BIT_HIGH;
	}
	else if (state == ZAC_READING_START_BIT_HIGH) {
		// rising edge of the start bit
		// at this stage the first rising edge of the start bit happened
		// we now have the strobe time acquired
		timer1_stop();
		strobe = TCNT1;
		// select falling edge, thus the startbit is complete
		// on next falling edge
		int1_switch_falling_edge();
		// next stage
		state = ZAC_ACQUIRING_TSTROBE;
	}
	else if (state == ZAC_ACQUIRING_TSTROBE) {
		// falling edge of start bit occured, reset timer count
		TCNT1 = 0;
		// start timer 1 in ctc mode, timer interrupt then occurs
		// each tstrobe/2 (50% duty cycle)
		timer1_start_ctc(strobe);
		// we now start reading the first byte
		state = ZAC_READING_HIGH_BYTE;
	}
	// now we are reading either the MSB or the LSB, start timer1 in ctc mode each falling edge
	else if (state == ZAC_READING_HIGH_BYTE || state == ZAC_READING_LOW_BYTE) {
		TCNT1 = 0;
		timer1_start_ctc(strobe);
	}
}

int main(void) {
	// PORTC all pins as output
	DDRC = 0xFF;
	int1_enable();
	int1_select_falling_edge();
	// turn TSIC on
	zac_on();
	_delay_us(125);
	// enable interrupts
	sei();
	// power on lcd
	lcd_init(LCD_DISP_ON);
	lcd_clrscr();
	// buffer for display
	char disp[16];
	// for temperature calculations
	uint16_t temperature = 0;
	float value = 0.0f;
	// main loop
	for(;;) {
		temperature = (buffer[0] << 8) | buffer [1];
		value = ((float)temperature / 2047 * 200) - 50;
		sprintf(disp,"%2.1f\337C  ",value);
		lcd_gotoxy(0,0);
		lcd_puts("Current temperature:");
		lcd_gotoxy(0,1);
		lcd_puts(disp);
		_delay_ms(100);
	}
	// never happens, just to make the compiler happy
	return 0;
}
